function heat(o)
% heat equation
% by Christian Himpe, 2014 ( http://gramian.de )
% released under BSD 2-Clause License ( opensource.org/licenses/BSD-2-Clause )
%*

if(exist('emgr')~=2) disp('emgr framework is required. Download at http://gramian.de/emgr.m'); return; end

rand('seed',1009); randn('seed',1009);

%%%%%%%% Setup %%%%%%%%

 n = 8;

 J = 1;
 N = n*n;
 O = J;
 T = [0 0.01 1.0];
 L = (T(3)-T(1))/T(2);
 U = [ones(J,1),zeros(J,L-1)];
 X =  zeros(N,1);

 %Discretization

 DSC_I = speye(n);
 DSC_E = ones(n,1);
 DSC_A = spdiags([DSC_E,-4.0*DSC_E,DSC_E],[-1,0,1],n,n);
 DSC_J = spdiags([DSC_E,DSC_E],[-1,1],n,n);
 A = (0.25*N)*(kron(DSC_I,DSC_A) + kron(DSC_J,DSC_I));

 P = 0.35*(rand(N,1)-0.5);

 B = sparse(N,1); B(1:n:end,1) = N*rand(n,1);
 C = sparse(1,N); C(1,n:n:end) = 1000*rand(n,1);

 LIN = @(x,u,p) (A+spdiags([0.25*p,0.25*p,p,0.25*p,0.25*p],[-n,-1 0 1,n],N,N))*x + B*u;
 OUT = @(x,u,p) C*x;
 ADJ = @(x,u,p) (A+spdiags([0.25*p,0.25*p,p,0.25*p,0.25*p],[-n,-1 0 1,n],N,N))'*x +C'*u;
 UNT = @(x,u,p) x;

 OT = 0;
 ZT = zeros(4,1);
 ZE = zeros(4,N);
 ZS = zeros(4,N);

%%%%%%%% Reduction %%%%%%%%%

% FULL
 tic;
 Y = ode(LIN,OUT,T,X,U,P);
 Z = ode(LIN,UNT,T,X,U,P);
 OT = toc;
 fprintf('|');

% OFFLINE  BT

%
% OFFLINE  WX
 tic; 
 WX = emgr(LIN,OUT,[J N O],T,'x',P,[0 0 0 0 0 0 0 0 0 0 0 -1]);
 [U0 D0 V0] = svd(WX);
 ZT(1,1) = toc;
 ZS(1,:) = diag(D0);
 fprintf('|');
%

% OFFLINE  WY
 tic; 
 WY = emgr(LIN,ADJ,[J N O],T,'y',P,[0 0 0 0 0 0 0 0 0 0 0 -1]);
 [U1 D1 V1] = svd(WY); 
 ZT(2,1) = toc;
 ZS(1,:) = diag(D1);
 fprintf('|');

% OFFLINE  WY(RESORTED)
 tic;
 [U3,D3,V3] = resort(A,B,C,U1,D1,U1');
 ZT(4,1) = toc;
 ZS(1,:) = diag(D3);
 fprintf('|');

%
% OFFLINE  WJ
 tic; 
 WJ = emgr(LIN,OUT,[J N O],T,'j',P,[0 0 0 0 0 0 0 0 0 0 0 -1]);
 [U4 D4 V4] = svd(WJ{1});
 [U5,D5,V5] = svd(WJ{2});
 ZT(5,1) = toc;
 fprintf('|\n');
%

 norm2 = @(y) sqrt(T(2)*sum(sum(y.*y)));

 for R=1:N

%
% ONLINE  WX
 [lin,out,x] = project(A,B,C,X,U0,U0',R);
 y = ode(lin,out,T,x,U,P);
 ZE(1,R) = norm2(Y - y)/norm2(Y);
%

% ONLINE  WY
 [lin,out,x] = project(A,B,C,X,U1,U1',R);
 y = ode(lin,out,T,x,U,P);
 yr = y;
 ZE(2,R) = norm2(Y - y)/norm2(Y);


% ONLINE  WY(RESORTED)
 [lin,out,x] = project(A,B,C,X,U3,U3',R);
 y = ode(lin,out,T,x,U,P);
 ZE(4,R) = norm2(Y - y)/norm2(Y);
%

% ONLINE  WJ(Combined)
 [lin,out,x] = project(A,B,C,X,U4,U4',R);
 y = ode(lin,out,T,x,U,U5(:,1:R)*U5(:,1:R)'*P);
 ZE(6,R) = norm2(Y - y)/norm2(Y);
%

 fprintf('|');
 if(mod(R,n)==0), fprintf('\n'); end;

 end

 fprintf('\n');

 ZT

figure;
semilogy(1:N,ZE(1,:),'r');
hold on;
semilogy(1:N,ZE(2,:),'b');
hold off;
xlim([1,12]); ylim([1e-3 1]);
legend('W_X','W_Y');
ylabel('Relative L^2 Output Error');
xlabel('State Dimension');
print('-depsc','wxwy');

figure;
semilogy(1:N,ZE(1,:),'r');
hold on;
semilogy(1:N,ZE(4,:),'k');
hold off;
xlim([1,12]); ylim([1e-3 1]);
legend('W_X','W_Y(RESORT)');
ylabel('Relative L^2 Output Error');
xlabel('State Dimension');
print('-depsc','wxwyre');

figure;
semilogy(1:N,ZE(1,:),'r');
hold on;
semilogy(1:N,ZE(6,:),'k');
hold off;
xlim([1,N]); ylim([1e-4 1]);
legend('W_X','W_J(COMBINED)');
ylabel('Relative L^2 Output Error');
xlabel('State Dimension');
print('-depsc','wxwj');

l = (1:-0.01:0)'; cmap = [l,l,ones(101,1)]; cmax = max(max([Y,y]));

figure;
plot(T(2):T(2):T(3),Y);
xlabel('Time');
ylabel('Output');

global ZA;

ZA = zeros(n,n,L);

for I=1:L
 ZA(:,:,I) = reshape(Z(:,I),[n,n]);
end

end

%%%%%%%% Projector %%%%%%%%

function [lin,out,x] = project(A,B,C,X,U,V,R)

 N = size(A,1);
 n = sqrt(N);

 uu = U(:,1:R);
 vv = V(1:R,:);
 b = vv*B;
 c = C*uu;
 x = vv*X;
 lin = @(x,u,p) vv*((A+spdiags([0.25*p,0.25*p,p,0.25*p,0.25*p],[-n,-1 0 1,n],N,N))*(uu*x)) + b*u;
 out = @(x,u,p) c*x;

end

%%%%%%%% Resorter %%%%%%%%

function [u,d,v] = resort(A,B,C,U,D,V)

 d = (abs((V*B).*(C*U)')).*diag(D);
 %d = (abs((V*B).*(V*B))).*diag(D);
 [d IX] = sort(d,'descend');
 for r = 1:size(A,1)
   u(:,r) = U(:,IX(r));
   v(r,:) = V(IX(r),:);
 end

 d = diag(d);

end

%%%%%%%% Integrator %%%%%%%%

function y = ode(f,g,t,x,u,p)

 h = t(2);
 H = 1.0./h;
 T = round((t(3)-t(1))/h);

 y = g(deval(ode23(@(t,y) f(y,u(1.0+min(round(t*H),T-1)),p),[0 h*T],x),linspace(0,h*T,T)),0,0);

end
